"""
Linux Platform Module
"""
# Copyright (C) 2007 Invisigoth - See LICENSE file for details
import os
import sys
import signal
import struct
import logging
import binascii
import platform
import traceback

import envi.cli as e_cli
import envi.bits as e_bits
import envi.const as e_const

import vtrace
import vtrace.exc as v_exc
import vtrace.breakpoints as v_bp

import vtrace.archs.arm as v_arm
import vtrace.archs.i386 as v_i386
import vtrace.archs.amd64 as v_amd64

import vtrace.platforms.base as v_base
import vtrace.platforms.posix as v_posix

from ctypes import *
import ctypes.util as cutil

logger = logging.getLogger(__name__)

if os.getenv('ANDROID_ROOT'):
    libc = CDLL('/system/lib/libc.so')
else:
    libc = CDLL(cutil.find_library("c"), use_errno=True)

libc.lseek64.restype = c_ulonglong
libc.lseek64.argtypes = [c_uint, c_ulonglong, c_uint]
libc.read.restype = c_long
libc.read.argtypes = [c_uint, c_void_p, c_long]
libc.write.restype = c_long
libc.write.argtypes = [c_uint, c_void_p, c_long]

O_RDWR = 2
O_LARGEFILE = 0x8000

MAP_ANONYMOUS = 0x20
MAP_PRIVATE = 0x02

# Linux specific ptrace extensions
PT_GETREGS = 12
PT_SETREGS = 13
PT_GETFPREGS = 14
PT_SETFPREGS = 15
PT_ATTACH = 16
PT_DETACH = 17
PT_GETFPXREGS = 18
PT_SETFPXREGS = 19
PT_SYSCALL = 24
PT_SETOPTIONS = 0x4200
PT_GETEVENTMSG = 0x4201
PT_GETSIGINFO = 0x4202
PT_SETSIGINFO = 0x4203
PT_GETREGSET = 0x4204
PT_SETREGSET = 0x4205
PT_SEIZE = 0x4206
PT_INTERRUPT = 0x4207
# PT set options stuff.  ONLY TRACESYSGOOD may be used in 2.4...
PT_O_TRACESYSGOOD   = 0x00000001 # add 0x80 to TRAP when generated by syscall
# For each of the options below, the stop signal is (TRAP | PT_EVENT_FOO << 8)
PT_O_TRACEFORK      = 0x00000002 # Cause a trap at fork
PT_O_TRACEVFORK     = 0x00000004 # Cause a trap at vfork
PT_O_TRACECLONE     = 0x00000008 # Cause a trap at clone
PT_O_TRACEEXEC      = 0x00000010 # Cause a trap at exec
PT_O_TRACEVFORKDONE = 0x00000020 # Cause a trap when vfork done
PT_O_TRACEEXIT      = 0x00000040 # Cause a trap on exit
PT_O_MASK           = 0x0000007f
# Ptrace event types (TRAP | PT_EVENT_FOO << 8) means that type
# when using GETEVENTMSG for most of these, the new pid is the data
PT_EVENT_FORK       = 1
PT_EVENT_VFORK      = 2
PT_EVENT_CLONE      = 3
PT_EVENT_EXEC       = 4
PT_EVENT_VFORK_DONE = 5
PT_EVENT_EXIT       = 6


# Reg sets for use in ptrace PT_GETREGSET
# typlically defined in /usr/include/elf.h
# for now, we only need the one, but we can add more on demand
NT_X86_XSTATE = 0x202  # x86 extended state using xsave

# Used to tell some of the additional events apart
SIG_LINUX_SYSCALL = signal.SIGTRAP | 0x80
SIG_LINUX_CLONE = signal.SIGTRAP | (PT_EVENT_CLONE << 8)
SIG_LINUX_EXIT = signal.SIGTRAP | (PT_EVENT_EXIT << 8)

#following from Pandaboard ES (OMAP4460) Armv7a (cortex-a9)
class user_regs_arm(Structure):
    _fields_ = (
            ("r0", c_ulong),
            ("r1", c_ulong),
            ("r2", c_ulong),
            ("r3", c_ulong),
            ("r4", c_ulong),
            ("r5", c_ulong),
            ("r6", c_ulong),
            ("r7", c_ulong),
            ("r8", c_ulong),
            ("r9", c_ulong),
            ("r10", c_ulong), #aka 'sl' ?
            ("r11", c_ulong),
            ("r12", c_ulong),
            ("sp", c_ulong),
            ("lr", c_ulong),
            ("pc", c_ulong),
            ("cpsr", c_ulong),
            ("orig_r0", c_ulong),
    )

class fp_reg_arm(Structure):
    _fields_ = (
            ("sign1", c_long, 1),
            ("unused", c_long, 15),
            ("sign2", c_long, 1),
            ("exponent", c_long, 14),
            ("j", c_long, 1),
            ("mantissa1", c_long, 31),
            ("mantissa0", c_long, 32),
    )

class user_fpregs_arm(Structure):
    _fields_ = (
            ("fpregs", fp_reg_arm*8),
            ("fpsr", c_ulong, 32),
            ("fpcr", c_ulong, 32),
            ("ftype", c_ubyte*8),
            ("init_flag", c_ulong),
    )

class USER_arm(Structure):
    _fields_ = (
        ("regs",       user_regs_arm),
        ("u_fpvalid",  c_long),
        ("u_tsize",    c_ulong),
        ("u_dsize",    c_ulong),
        ("u_ssize",    c_ulong),
        ("start_code", c_ulong),
        ("start_stack",c_ulong),
        ("signal",     c_long),
        ("reserved",   c_long),
        ("u_ar0",      c_void_p),
        ("magic",      c_ulong),
        ("u_comm",     c_char*32),
        ("u_debugreg", c_long*8),
        ("fpregs",     user_fpregs_arm),
        ("u_fp0",      c_void_p)
    )

class user_regs_i386(Structure):
    _fields_ = (
        ("ebx",  c_ulong),
        ("ecx",  c_ulong),
        ("edx",  c_ulong),
        ("esi",  c_ulong),
        ("edi",  c_ulong),
        ("ebp",  c_ulong),
        ("eax",  c_ulong),
        ("ds",   c_ushort),
        ("__ds", c_ushort),
        ("es",   c_ushort),
        ("__es", c_ushort),
        ("fs",   c_ushort),
        ("__fs", c_ushort),
        ("gs",   c_ushort),
        ("__gs", c_ushort),
        ("orig_eax", c_ulong),
        ("eip",  c_ulong),
        ("cs",   c_ushort),
        ("__cs", c_ushort),
        ("eflags", c_ulong),
        ("esp",  c_ulong),
        ("ss",   c_ushort),
        ("__ss", c_ushort),
    )

class iovec(Structure):
    _fields_ = [
        ('iov_base', c_void_p),
        ('iov_len', c_size_t),
    ]

class user_fpregs_i386(Structure):
    _fields_ = [
        ('cwd', c_long),
        ('swd', c_long),
        ('twd', c_long),
        ('fip', c_long),
        ('fcs', c_long),
        ('foo', c_long),
        ('fos', c_long),
        ('st_space', c_long * 20),
    ]


class user_fpxregs_i386(Structure):
    _fields_ = [
        ('cwd', c_ushort),
        ('swd', c_ushort),
        ('twd', c_ushort),
        ('fop', c_ushort),
        ('fip', c_long),
        ('fcs', c_long),
        ('foo', c_long),
        ('fos', c_long),
        ('mxcsr', c_long),
        ('reserved', c_long),
        ('st_space', c_long * 32),
        ('xmm_space', c_long * 32),
        ('padding', c_long * 56),
    ]


class USER_i386(Structure):
    _fields_ = (
        # NOTE: Expand out the user regs struct so
        #       we can make one call to _rctx_Import
        ("regs",       user_regs_i386),
        ("u_fpvalid",  c_ulong),
        ("u_tsize",    c_ulong),
        ("u_dsize",    c_ulong),
        ("u_ssize",    c_ulong),
        ("start_code", c_ulong),
        ("start_stack",c_ulong),
        ("signal",     c_ulong),
        ("reserved",   c_ulong),
        ("u_ar0",      c_void_p),
        ("u_fpstate",  c_void_p),
        ("magic",      c_ulong),
        ("u_comm",     c_char*32),
        ("debug0",     c_ulong),
        ("debug1",     c_ulong),
        ("debug2",     c_ulong),
        ("debug3",     c_ulong),
        ("debug4",     c_ulong),
        ("debug5",     c_ulong),
        ("debug6",     c_ulong),
        ("debug7",     c_ulong),
    )

class user_regs_amd64(Structure):
    _fields_ = [
        ('r15',      c_uint64),
        ('r14',      c_uint64),
        ('r13',      c_uint64),
        ('r12',      c_uint64),
        ('rbp',      c_uint64),
        ('rbx',      c_uint64),
        ('r11',      c_uint64),
        ('r10',      c_uint64),
        ('r9',       c_uint64),
        ('r8',       c_uint64),
        ('rax',      c_uint64),
        ('rcx',      c_uint64),
        ('rdx',      c_uint64),
        ('rsi',      c_uint64),
        ('rdi',      c_uint64),
        ('orig_rax', c_uint64),
        ('rip',      c_uint64),
        ('cs',       c_uint64),
        ('eflags',   c_uint64),
        ('rsp',      c_uint64),
        ('ss',       c_uint64),
        ('fs_base',  c_uint64),
        ('gs_base',  c_uint64),
        ('ds',       c_uint64),
        ('es',       c_uint64),
        ('fs',       c_uint64),
        ('gs',       c_uint64),
    ]


intel_dbgregs = (0,1,2,3,6,7)


class LinuxMixin(v_posix.PtraceMixin, v_posix.PosixMixin):
    """
    The mixin to take care of linux specific platform traits.
    (mostly proc)
    """

    def __init__(self):
        # Wrap reads from proc in our worker thread
        v_posix.PtraceMixin.__init__(self)
        v_posix.PosixMixin.__init__(self)
        self.memfd = None
        self._stopped_cache = {}
        self._stopped_hack = False

        self.fireTracerThread()
        self.setMeta('BadMaps', ['[vvar]', '[vsyscall]', '[vvar_vclock]'])

        self.initMode("Syscall", False, "Break On Syscalls")

    def setupMemFile(self, offset):
        """
        A utility to open (if necessary) and seek the memfile
        """
        if self.memfd is None:
            self.memfd = libc.open(b"/proc/%d/mem" % self.pid, O_RDWR | O_LARGEFILE, 0o755)
            if self.memfd < 0:
                logger.warning('Failed to get proper file descriptor (errno: %d)', get_errno())

        retn = libc.lseek64(self.memfd, offset, 0)
        if retn < 0:
            logger.warning('lseek64 hit issue with error: %d' % get_errno())

    @v_base.threadwrap
    def platformReadMemory(self, address, size):
        """
        A *much* faster way of reading memory that the 4 bytes
        per syscall allowed by ptrace
        """
        self.setupMemFile(address)
        # Use ctypes cause python implementation is teh ghey
        buf = create_string_buffer(size)
        x = libc.read(self.memfd, addressof(buf), size)
        if x != size:
            # libc.perror('libc.read %d (size: %d)' % (x,size))
            raise Exception("reading from invalid memory %s (%d returned) (errno: %d) (fd: %d)" % (hex(address), x, get_errno(), self.memfd))
        # We have to slice cause ctypes "helps" us by adding a null byte...
        return buf.raw

    @v_base.threadwrap
    def whynot_platformWriteMemory(self, address, data):
        """
        A *much* faster way of writting memory that the 4 bytes
        per syscall allowed by ptrace
        """
        self.setupMemFile(address)
        buf = create_string_buffer(data)
        size = len(data)
        x = libc.write(self.memfd, addressof(buf), size)
        if x != size:
            libc.perror('write mem failed: 0x%.8x (%d)' % (address, size))
            raise Exception("write memory failed: %d" % x)
        return x

    def _findExe(self, pid):
        exe = os.readlink("/proc/%d/exe" % pid)
        if "(deleted)" in exe:
            if "#prelink#" in exe:
                exe = exe.split(".#prelink#")[0]
            elif ";" in exe:
                exe = exe.split(";")[0]
            else:
                exe = exe.split("(deleted)")[0].strip()
        return exe

    @v_base.threadwrap
    def platformExec(self, cmdline):
        # Very similar to posix, but not
        # quite close enough...
        self.execing = True
        cmdlist = e_cli.splitargs(cmdline)
        os.stat(cmdlist[0])
        pid = os.fork()

        if pid == 0:
            try:
                # Don't use PT_TRACEME -- on some linux (tested on ubuntu)
                # it will cause immediate asignment of ptrace slot to parent
                # without parent having PT_ATTACH'D.... MAKES SYNCHRONIZATION HARD
                # SIGSTOP ourself until parent continues us
                os.kill(os.getpid(), signal.SIGSTOP)
                os.execv(cmdlist[0], cmdlist)
            except Exception as e:
                logger.error(e)
            sys.exit(-1)

        # Attach to child. should cause SIGSTOP
        if 0 != v_posix.ptrace(PT_ATTACH, pid, 0, 0):
            raise Exception("PT_ATTACH failed! linux platformExec")

        # Eat all SIGSTOP (or other signal) and break from loop on SIGTRAP.
        # SIGTRAP triggered by execv while PTRACE_ATTACH'd
        while True:
            wpid, status = os.waitpid(pid, os.WUNTRACED)
            if wpid != pid:  # should never happen
                continue
            if os.WIFSTOPPED(status):
                cause = os.WSTOPSIG(status)
                if cause == signal.SIGTRAP:
                    break
                if v_posix.ptrace(v_posix.PT_CONTINUE, pid, 0, 0) != 0:
                    raise Exception("PT_CONTINUE failed! linux platformExec")

        # Do a single step, which will allow a new stop event for the
        # rest of vtrace to eat up.
        if v_posix.ptrace(v_posix.PT_STEP, pid, 0, 0) != 0:
            raise Exception("PT_CONTINUE failed! linux platformExec")

        self.pthreads = [pid]
        self.setMeta("ExeName", self._findExe(pid))
        return pid

    @v_base.threadwrap
    def platformAttach(self, pid):
        self.pthreads = [pid]
        self.setMeta("ThreadId", pid)
        if v_posix.ptrace(PT_ATTACH, pid, 0, 0) != 0:
            raise Exception("PT_ATTACH failed!")
        self.setMeta("ExeName", self._findExe(pid))

    def _LibraryLoadHook(self):
        # drop special breakpoint at ld._dl_catch_exception
        bp = v_bp.PosixLibLoadHookBreakpoint('ld._dl_catch_exception')
        self.addBreakpoint(bp)

    def platformPs(self):
        pslist = []
        for dname in self.platformListDir('/proc'):
            try:
                if not dname.isdigit():
                    continue
                cmdline = self.platformReadFile('/proc/%s/cmdline' % dname)
                cmdline = cmdline.replace(b"\x00", b" ")
                if len(cmdline) > 0:
                    pslist.append((int(dname), cmdline.decode('utf-8')))
            except Exception as e:
                pass  # Permissions...  quick process... whatev.
        return pslist

    def _simpleCreateThreads(self):
        for tid in self.threadsForPid( self.pid ):
            if tid == self.pid:
                continue
            self.attachThread( tid )

    def attachThread(self, tid, attached=False):
        self.doAttachThread(tid, attached=attached)
        self.setMeta("ThreadId", tid)
        self.fireNotifiers(vtrace.NOTIFY_CREATE_THREAD)

    @v_base.threadwrap
    def detachThread(self, tid, ecode):

        self.setMeta('ThreadId', tid)
        self._fireExitThread(tid, ecode)

        if v_posix.ptrace(PT_DETACH, tid, 0, 0) != 0:
            raise Exception("ERROR ptrace detach failed for thread %d" % tid)

        self.pthreads.remove(tid)

    @v_base.threadwrap
    def platformWait(self):
        # Blocking wait once...
        pid, status = os.waitpid(-1, 0x40000002)
        self.setMeta("ThreadId", pid)
        # Stop the rest of the threads...
        # why is linux debugging so Ghetto?!?!
        if not self.stepping:  # If we're stepping, only do the one
            for tid in self.pthreads:
                if tid == pid:
                    continue
                try:
                    # We use SIGSTOP here because they can't mask it.
                    os.kill(tid, signal.SIGSTOP)
                    os.waitpid(tid, 0x40000002)
                except Exception as e:
                    logger.warning("WARNING TID is invalid %d %s", tid, e)
        return pid, status

    @v_base.threadwrap
    def platformContinue(self):
        cmd = v_posix.PT_CONTINUE
        if self.getMode("Syscall", False):
            cmd = PT_SYSCALL
        pid = self.getPid()
        sig = self.getCurrentSignal()
        if sig is None:
            sig = 0
        # Only deliver signals to the main thread
        if v_posix.ptrace(cmd, pid, 0, sig) != 0:
            libc.perror('ptrace PT_CONTINUE failed for pid %d' % pid)
            raise Exception("ERROR ptrace failed for pid %d" % pid)

        for tid in self.pthreads:
            if tid == pid:
                continue
            if v_posix.ptrace(cmd, tid, 0, 0) != 0:
                pass

    @v_base.threadwrap
    def platformStepi(self):
        self.stepping = True
        tid = self.getMeta("ThreadId", 0)
        if v_posix.ptrace(v_posix.PT_STEP, tid, 0, 0) != 0:
            raise Exception("ERROR ptrace failed!")

    @v_base.threadwrap
    def platformDetach(self):
        libc.close(self.memfd)
        for tid in self.pthreads:
            v_posix.ptrace(PT_DETACH, tid, 0, 0)

    @v_base.threadwrap
    def doAttachThread(self, tid, attached=False):
        """
        Do the work for attaching a thread.  This must be *under*
        attachThread() so callers in notifiers may call it (because
        it's also gotta be thread wrapped).
        """
        if not attached:
            if v_posix.ptrace(PT_ATTACH, tid, 0, 0) != 0:
                raise Exception("ERROR ptrace attach failed for thread %d" % tid)

        # We may have already revcieved the stop signal
        if not self._stopped_cache.pop(tid, None):
            os.waitpid(tid, 0x40000002)

        self.setupPtraceOptions(tid)
        self.pthreads.append(tid)

    @v_base.threadwrap
    def setupPtraceOptions(self, tid):
        """
        Called per pid/tid to setup proper options
        for ptrace.
        """
        opts = PT_O_TRACESYSGOOD
        ver = tuple(platform.release()[:3].split('.'))
        if (int(ver[0]), int(ver[1])) >= (2, 6):
            opts |= PT_O_TRACECLONE | PT_O_TRACEEXIT
        x = v_posix.ptrace(PT_SETOPTIONS, tid, 0, opts)
        if x != 0:
            libc.perror('ptrace PT_SETOPTION failed for thread %d' % tid)

    def threadsForPid(self, pid):
        ret = []
        tpath = "/proc/%s/task" % pid
        if os.path.exists(tpath):
            for pidstr in os.listdir(tpath):
                ret.append(int(pidstr))
        return ret

    def platformProcessEvent(self, event):
        # Skim some linux specific events before passing to posix
        tid, status = event
        if os.WIFSTOPPED(status):
            sig = status >> 8 # Cant use os.WSTOPSIG() here...
            # print('STOPPED: %d %d %.8x %d' % (self.pid, tid, status, sig))

            # Ok... this is a crazy little state engine that tries
            # to account for the discrepancies in how linux posts
            # signals to the debugger...

            # Thread Creation:
            # In each case below, the kernel may deliver
            # any of the 3 signals in any order...  ALSO
            # (and more importantly) *if* the kernel sends
            # SIGSTOP to the thread first, the debugger
            # will get a SIGSTOP *instead* of SIG_LINUX_CLONE
            # ( this will go back and forth on *ONE BOX* with
            #   the same kernel version... Finally squished it
            #   because it presents more frequently ( 1 in 10 )
            #   on my new ARM linux dev board. WTF?!1?!one?!? )
            #
            # Case 1 (SIG_LINUX_CLONE):
            #     debugger gets SIG_LINUX CLONE as expected
            #     and can then use ptrace(PT_GETEVENTMSG)
            #     to get new TID and attach as normal
            # Case 2 (SIGSTOP delivered to thread)
            #     Thread is already stoped and attached but
            #     parent debugger doesn't know yet.  We add
            #     the tid to the stopped_cache so when the
            #     kernel gets around to telling the debugger
            #     we don't wait on him again.
            # Case 3 (SIGSTOP delivered to debugger)
            #     In both case 2 and case 3, this will cause
            #     the SIG_LINUX_CLONE to be skipped.  Either
            #     way, we should head down into thread attach.
            #     ( The thread may be already stopped )
            if sig == SIG_LINUX_SYSCALL:
                self.fireNotifiers(vtrace.NOTIFY_SYSCALL)

            elif sig == SIG_LINUX_EXIT:

                ecode = self.getPtraceEvent() >> 8

                if tid == self.getPid():
                    self._fireExit( ecode )
                    self.platformDetach()

                else:
                    self.detachThread(tid, ecode)

            elif sig == SIG_LINUX_CLONE:
                # Handle a new thread here!
                newtid = self.getPtraceEvent()
                # print('CLONE (new tid: %d)' % newtid)
                self.attachThread(newtid, attached=True)

            elif sig == signal.SIGSTOP and tid != self.pid:
                #print('OMG IM THE NEW THREAD! %d' % tid)
                # We're not even a real event right now...
                self.runAgain()
                self._stopped_cache[tid] = True

            elif sig == signal.SIGSTOP:
                # If we are still 'exec()'ing, we havent hit the SIGTRAP
                # yet ( so our process info is still python, lets skip it )
                if self.execing:
                    self._stopped_hack = True
                    self.setupPtraceOptions(tid)
                    self.runAgain()

                elif self._stopped_hack:
                    newtid = self.getPtraceEvent(tid)
                    #print("WHY DID WE GET *ANOTHER* STOP?: %d" % tid)
                    #print('PTRACE EVENT: %d' % newtid)
                    self.attachThread(newtid, attached=True)

                else: # on first attach...
                    self._stopped_hack = True
                    self.setupPtraceOptions(tid)
                    self.handlePosixSignal(sig)

            #FIXME eventually implement child catching!
            else:
                self.handlePosixSignal(sig)

            return

        v_posix.PosixMixin.platformProcessEvent(self, event)

    @v_base.threadwrap
    def getPtraceEvent(self, tid=None):
        """
        This *thread wrapped* function will get any pending GETEVENTMSG
        msgs.
        """
        p = c_ulong(0)
        if tid is None:
            tid = self.getMeta("ThreadId", -1)
        if v_posix.ptrace(PT_GETEVENTMSG, tid, 0, addressof(p)) != 0:
            raise Exception('ptrace PT_GETEVENTMSG failed!')
        return p.value

    def platformGetThreads(self):
        ret = {}
        for tid in self.pthreads:
            ret[tid] = tid  # FIXME make this pthread struct or stackbase soon
        return ret

    def platformGetMaps(self):
        maps = []
        with open("/proc/%d/maps" % self.pid, 'r') as mapfile:
            for line in mapfile:

                perms = 0
                sline = line.split(" ")
                addrs = sline[0]
                permstr = sline[1]
                fname = sline[-1].strip()
                addrs = addrs.split("-")
                base = int(addrs[0],16)
                max = int(addrs[1],16)
                mlen = max-base

                if "r" in permstr:
                    perms |= e_const.MM_READ
                if "w" in permstr:
                    perms |= e_const.MM_WRITE
                if "x" in permstr:
                    perms |= e_const.MM_EXEC
                #if "p" in permstr:
                    #pass

                maps.append((base,mlen,perms,fname))
        return maps

    def platformGetFds(self):
        fds = []
        for name in os.listdir("/proc/%d/fd/" % self.pid):
            try:
                fdnum = int(name)
                fdtype = vtrace.FD_UNKNOWN
                link = os.readlink("/proc/%d/fd/%s" % (self.pid, name))
                if "socket:" in link:
                    fdtype = vtrace.FD_SOCKET
                elif "pipe:" in link:
                    fdtype = vtrace.FD_PIPE
                elif "/" in link:
                    fdtype = vtrace.FD_FILE

                fds.append((fdnum, fdtype, link))
            except Exception:
                logger.error(traceback.format_exc())

        return fds

############################################################################
#
# NOTE: Both of these use class locals set by the i386/amd64 variants
#
    @v_base.threadwrap
    def platformGetRegCtx(self, tid):
        ctx = self.archGetRegCtx()
        u = self.user_reg_struct()
        if v_posix.ptrace(PT_GETREGS, tid, 0, addressof(u)) == -1:
            raise v_exc.PtraceException("PT_GETREGS")

        ctx._rctx_Import(u)
        return ctx

    @v_base.threadwrap
    def platformSetRegCtx(self, tid, ctx):
        u = self.user_reg_struct()
        # Populate the reg struct with the current values (to allow for
        # any regs in that struct that we don't track... *fs_base*ahem*
        if v_posix.ptrace(PT_GETREGS, tid, 0, addressof(u)) == -1:
            raise v_exc.PtraceException("PT_GETREGS")

        ctx._rctx_Export(u)
        if v_posix.ptrace(PT_SETREGS, tid, 0, addressof(u)) == -1:
            raise v_exc.PtraceException("PT_SETREGS")

class Linuxi386Trace(
        vtrace.Trace,
        LinuxMixin,
        v_i386.i386Mixin,
        v_posix.ElfMixin,
        v_base.TracerBase):

    user_reg_struct = user_regs_i386
    user_dbg_offset = 252
    reg_val_mask = 0xffffffff

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_posix.ElfMixin.__init__(self)
        v_i386.i386Mixin.__init__(self)
        LinuxMixin.__init__(self)

        # Pre-calc the index of the debug regs
        self.dbgidx = self.archGetRegCtx().getRegisterIndex("debug0")

    @v_base.threadwrap
    def platformGetRegCtx(self, tid):
        ctx = LinuxMixin.platformGetRegCtx( self, tid )
        for i in intel_dbgregs:
            offset = self.user_dbg_offset + (self.psize * i)
            r = v_posix.ptrace(v_posix.PT_READ_U, tid, offset, 0)
            ctx.setRegister(self.dbgidx+i, r & self.reg_val_mask)
        return ctx

    @v_base.threadwrap
    def platformSetRegCtx(self, tid, ctx):
        LinuxMixin.platformSetRegCtx( self, tid, ctx )
        for i in intel_dbgregs:
            val = ctx.getRegister(self.dbgidx + i)
            offset = self.user_dbg_offset + (self.psize * i)
            if v_posix.ptrace(v_posix.PT_WRITE_U, tid, offset, val) != 0:
                libc.perror('PT_WRITE_U failed for debug%d' % i)

    @v_base.threadwrap
    def platformAllocateMemory(self, size, perms=e_const.MM_RWX, suggestaddr=0):
        sp = self.getStackCounter()
        pc = self.getProgramCounter()

        # Xlate perms (mmap is backward)
        realperm = 0
        if perms & e_const.MM_READ:
            realperm |= 1
        if perms & e_const.MM_WRITE:
            realperm |= 2
        if perms & e_const.MM_EXEC:
            realperm |= 4

        #mma is struct of mmap args for linux syscall
        mma = struct.pack("<6L", suggestaddr, size, realperm, MAP_ANONYMOUS|MAP_PRIVATE, 0, 0)

        regsave = self.getRegisters()

        stacksave = self.readMemory(sp, len(mma))
        ipsave = self.readMemory(pc, 2)

        SYS_mmap = 90

        self.writeMemory(sp, mma)
        self.writeMemory(pc, b"\xcd\x80")
        self.setRegisterByName("eax", SYS_mmap)
        self.setRegisterByName("ebx", sp)
        self._syncRegs()

        try:
            # Step over our syscall instruction
            tid = self.getMeta("ThreadId", 0)
            self.platformStepi()
            os.waitpid(tid, 0)
            eax = self.getRegisterByName("eax")
            if eax & 0x80000000:
                raise Exception("Linux mmap syscall error: %d" % eax)
            return eax

        finally:
            # Clean up all our fux0ring
            self.writeMemory(sp, stacksave)
            self.writeMemory(pc, ipsave)
            self.setRegisters(regsave)


class LinuxAmd64Trace(
        vtrace.Trace,
        LinuxMixin,
        v_amd64.Amd64Mixin,
        v_posix.ElfMixin,
        v_base.TracerBase):

    user_reg_struct = user_regs_amd64
    user_dbg_offset = 848
    reg_val_mask = 0xffffffffffffffff

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_posix.ElfMixin.__init__(self)
        v_amd64.Amd64Mixin.__init__(self)
        LinuxMixin.__init__(self)
        self.dbgidx = self.archGetRegCtx().getRegisterIndex("debug0")

    @v_base.threadwrap
    def platformGetRegCtx(self, tid):
        ctx = LinuxMixin.platformGetRegCtx( self, tid )
        for i in intel_dbgregs:
            offset = self.user_dbg_offset + (self.psize * i)
            r = v_posix.ptrace(v_posix.PT_READ_U, tid, offset, 0)
            ctx.setRegister(self.dbgidx+i, r & self.reg_val_mask)

        self.platformGetExtendedRegs(tid, ctx)
        return ctx

    @v_base.threadwrap
    def platformSetRegCtx(self, tid, ctx):
        LinuxMixin.platformSetRegCtx( self, tid, ctx )
        for i in intel_dbgregs:
            val = ctx.getRegister(self.dbgidx + i)
            offset = self.user_dbg_offset + (self.psize * i)
            if v_posix.ptrace(v_posix.PT_WRITE_U, tid, offset, val) != 0:
                libc.perror('PT_WRITE_U failed for debug%d' % i)

    def parseXSave(self, ctx, iovec):
        '''
        PT_GETREGSET basically just dumps the xsave memory region, so we need to parse that out
        '''
        # There are other registers that are saved in here, but for now, fuck 'em
        simd_regs = [0 for i in range(16)]
        fpu_off = 32
        fpu_len = 8
        # yes you could get this by various other PTRACE calls, but since we already have to call into NT_X86_XSTATE, 
        # might as well grab these while we're here
        regidx = self.archGetRegCtx().getRegisterIndex("st0")
        mask = 0xFFFFFFFFFFFFFFFFFFFF  # 80 bit mask for floating points regs
        for i in range(fpu_len):
            offset = fpu_off + i * 16
            # the upper 48 bits of the st/mm registers are marked as reserved
            valu = e_bits.parsebytes(bytes(iovec[offset:offset+10]), 0, 10)
            ctx.setRegister(regidx+i, valu)

        xmm_off = 160

        # these are just the lower bits of the simd registers (just the xmm portion)
        for i in range(len(simd_regs)):
            offset = xmm_off + i * 16
            simd_regs[i] = e_bits.parsebytes(bytes(iovec[offset:offset+16]), 0, 16)

        xstate_bv = e_bits.parsebytes(bytes(iovec[512:520]), 0, 8)
        has_avx = xstate_bv & 0x4
        # XXX: Sooooo....we're gonna cheat a bit here. Technically what we're supposed to do
        # is check CPUID.(EAX=0x0D, ECX=i) for every feature and se how many bytes it takes up, but 
        # right now the goal is just to get the upper YMM registers, and we know exactly how
        # many bytes those take up, and they're literally the first state component in the extended
        # xsave region (standard or compacted), so let's parse us some bytes if the has_avx bit is set.
        # (we're also doing it this way because I don't feel like figuring out how to directly call
        # the cpuid asm instruction from python)
        if has_avx:
            ymm_offset = 576
            for i in range(len(simd_regs)):
                offset = ymm_offset + i*16
                valu = e_bits.parsebytes(bytes(iovec[offset:offset+16]), 0, 16)
                simd_regs[i] |= valu << 128

        regidx = self.archGetRegCtx().getRegisterIndex("ymm0")
        for i, valu in enumerate(simd_regs):
            ctx.setRegister(regidx+i, valu)

    def platformGetExtendedRegs(self, tid, ctx):
        '''
        for now, the only real way to get access to things like YMM and ZMM registers
        '''
        buflen = 2048  # Guess. Actual length will be set by kernel
        buffer = (c_uint8 * buflen)()
        vec = iovec(cast(buffer, c_void_p), buflen)

        if v_posix.ptrace(PT_GETREGSET, tid, NT_X86_XSTATE, addressof(vec)) != 0:
            raise v_exc.PtraceException("PT_GETREGSET(NT_X86_XSTATE)")

        self.parseXSave(ctx, buffer)



arm_break_be = binascii.unhexlify('e7f001f0')
arm_break_le = binascii.unhexlify('f001f0e7')

class LinuxArmTrace(
        vtrace.Trace,
        LinuxMixin,
        v_arm.ArmMixin,
        v_posix.ElfMixin,
        v_base.TracerBase):

    user_reg_struct = user_regs_arm
    reg_val_mask = 0xffffffff

    def __init__(self):
        vtrace.Trace.__init__(self)
        v_base.TracerBase.__init__(self)
        v_posix.ElfMixin.__init__(self)
        v_arm.ArmMixin.__init__(self)
        LinuxMixin.__init__(self)

        self._break_after_bp = False
        self._step_cleanup = []

    def _fireStep(self):
        # See notes below about insanity...
        if self._step_cleanup is not None:
            [ self.writeMemory( bva, bytes ) for (bva,bytes) in self._step_cleanup ]
            self._step_cleanup = None
        return v_base.TracerBase._fireStep( self )

    def archGetBreakInstr(self):
        return arm_break_le

    @v_base.threadwrap
    def platformStepi(self):
        # This is a total rediculous hack to account
        # for the fact that the arm platform couldn't
        # be bothered to implement single stepping in
        # the stupid hardware...

        self.stepping = True

        pc = self.getProgramCounter()
        op = self.parseOpcode( pc )

        branches = op.getBranches( self )
        if not branches:
            raise Exception('''
                    The branches for the instruction %r were not decoded correctly.  This means that
                    we cant properly predict the possible next instruction executions in a way that allows us
                    to account for the STUPID INSANE FACT THAT THERE IS NO HARDWARE SINGLE STEP CAPABILITY ON
                    ARM (non-realtime or JTAG anyway).  We *would* have written invalid instructions to each
                    of those locations and cleaned them up before you ever knew anything was amiss... which is
                    how we pretend arm can single step... even though IT CANT. (please tell visi...)
            ''' % op)

        # Save the memory at the branches for later
        # restoration in the _fireStep callback.

        self._step_cleanup = []
        for bva,bflags in op.getBranches( self ):
            self._step_cleanup.append( (bva, self.readMemory( bva, 4 )) )
            self.writeMemory( bva, arm_break_le )

        tid = self.getMeta('ThreadId')

        if v_posix.ptrace(v_posix.PT_CONTINUE, tid, 0, 0) != 0:
            raise Exception("ERROR ptrace failed for tid %d" % tid)




import vivisect.exc as viv_exc


class VivColorMap:
    '''
    An easier to use color map object.
    '''

    def __init__(self, vw):
        self.vw = vw
        self.cmap = {}

    def colorVa(self, va, color):
        '''
        Set the color for a particular virtual memory location
        when viewed with this color map.

        Example: m.colorVa(va, 'red')
        '''
        self.cmap[va] = color

    def colorBlock(self, va, color):
        '''
        Similar to colorVa(), but loops coloring all the virtual
        addresses in the given block.
        '''
        cbtup = self.vw.getCodeBlock(va)
        if cbtup is None:
            raise viv_exc.InvalidCodeBlock(va)

        cbva, cbsize, fva = cbtup
        for i in range(cbsize):
            self.cmap[cbva + i] = color

    def colorFunction(self, fva, color):
        '''
        Similar to colorVa(), but loops coloring all the virtual
        addresses in the given function.
        '''
        fva = self.vw.getFunction(fva)
        for cbva, cbsize, fva in self.vw.getFunctionBlocks(fva):
            for i in range(cbsize):
                self.cmap[cbva + i] = color

    def saveAs(self, name):
        '''
        Save the color map to the workspace with the given name.
        '''
        self.vw.addColorMap(name, dict(self.cmap))

    def setGuiMap(self):
        '''
        Set this as the display color map for the vivisect gui.  (obviously
        only works when GUI is running...)
        '''
        self.vw._viv_gui.setColorMap(self.cmap)

    def getColorDict(self):
        return dict(self.cmap)



# Base Microsoft protocol, but many fields are left undocumented
# https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-SHLLINK/%5bMS-SHLLINK%5d.pdf

# Other research for filling in the things Microsoft didn't
# https://github.com/libyal/libfwsi/blob/main/documentation/Windows%20Shell%20Item%20format.asciidoc

# NOTE: all things are little-endian

import logging

import vstruct

from enum import IntFlag, IntEnum, auto

from vstruct.primitives import v_uint32, v_uint16, v_uint8, v_zstr, v_zwstr, GUID, v_uint64, v_wstr, v_str, v_bytes

from vstruct.defs.mspropstore import PropertyStorage

logger = logging.getLogger(__name__)


class LinkFlags(IntFlag):
    HasLinkTargetIDList = auto()
    HasLinkInfo = auto()
    HasName = auto()
    HasRelativePath = auto()
    HasWorkingDir = auto()
    HasArguments = auto()
    HasIconLocation = auto()
    IsUnicode = auto()
    ForceNoLinkInfo = auto()
    HasExpString = auto()
    RunInSeparateProcess = auto()
    Unused1 = auto()
    HasDawinID = auto()
    RunAsUser = auto()
    HasExpIcon = auto()
    NoPidAlias = auto()
    Unused2 = auto()
    RunWithShimLayer = auto()
    ForceNoLinkTrack = auto()
    EnableTargetMetadata = auto()
    DisableLinkPathTracking = auto()
    DisableKnownFolderTracking = auto()
    DisableKnownFolderAlias = auto()
    AllowLinkToLink = auto()
    UnaliasOnSave = auto()
    PreferEnvironmentPath = auto()
    KeepLocalIDListForUNCTarget = auto()


class LinkInfoFlags(IntFlag):
    VolumeIDAndLocalBasePath = auto()
    CommonNetworkRelativeLinkAndPathSuffix = auto()


class DriveTypes(IntEnum):
    Unknown = auto()
    NoRootDir = auto()
    Removable = auto()
    Fixed = auto()
    Remote = auto()
    Cdrom = auto()
    Ramdisk = auto()


class ExtraDataSigs(IntEnum):
    Console = 0xA0000002
    ConsoleFE = 0xA0000004
    Darwin = 0xA0000006
    Environment = 0xA0000001
    IconEnvironment = 0xA0000007
    KnownFolder = 0xA000000B
    PropertyStore = 0xA0000009
    Shim = 0xA0000008
    SpecialFolder = 0xA0000005
    Tracker = 0xA0000003
    VistaAndAboveIDList = 0xA000000C


# <Header> [<Link target Id LIST>] [<LinkInfo>] [<String data>] [<extra data>]
class LnkShellLinkHeader(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.headersize = v_uint32()
        self.linkclsid = GUID()
        self.linkflags = v_uint32()
        self.fileattrs = v_uint32()

        self.created = v_uint64()
        self.accessed = v_uint64()
        self.written = v_uint64()

        self.filesize = v_uint32()
        self.iconindex = v_uint32()
        self.showcmd = v_uint32()

        self.hotkey = v_uint16()

        self.reserved1 = v_uint16()
        self.reserved2 = v_uint32()
        self.reserved3 = v_uint32()


class VolumeID(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.size = v_uint32()
        self.drivetype = v_uint32()
        self.driveserial = v_uint32()
        self.volumelabeloffset = v_uint32()

    def pcb_volumelabeloffset(self):
        if self.volumelabeloffset == 0x14:
            self.volumelabeloffsetunicode = v_uint32()
            self.data = v_zwstr()
        else:
            self.data = v_zstr()


class ItemBase(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.size = v_uint16()
        self.iden = v_uint8()


class ExtraDataBlock(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.blocksize = v_uint32()
        self.blocksig = v_uint32()


class ConsoleBlock(ExtraDataBlock):
    def __init__(self):
        ExtraDataBlock.__init__(self)

        self.fillAttrs = v_uint16()
        self.popupFillAttrs = v_uint16()

        self.screenBufferSizeX = v_uint16()
        self.screenBufferSizeY = v_uint16()

        self.windowSizeX = v_uint16()
        self.windowSizeY = v_uint16()

        self.windowOriginX = v_uint16()
        self.windowOriginY = v_uint16()

        self.unused1 = v_uint32()
        self.unused2 = v_uint32()

        self.fontSize = v_uint32()
        self.fontFamily = v_uint32()
        self.fontWeight = v_uint32()

        self.faceName = v_wstr(size=32)  # unicode, 64 bytes

        self.cursorSize = v_uint32()
        self.fullScreen = v_uint32()
        self.quickEdit = v_uint32()

        self.insertMode = v_uint32()
        self.autoPosition = v_uint32()
        self.historyBufferSize = v_uint32()
        self.numberofHistoryBuffers = v_uint32()
        self.historyNoDup = v_uint32()
        self.colorTable = vstruct.VArray([v_uint32() for i in range(16)])


class ConsoleFEBlock(ExtraDataBlock):
    def __init__(self):
        ExtraDataBlock.__init__(self)

        self.codepage = v_uint32()


class TargetBlock(ExtraDataBlock):
    def __init__(self):
        ExtraDataBlock.__init__(self)
        self.targetAnsi = v_str(size=260)
        self.targetUnicode = v_wstr(size=520)


class FolderBlock(ExtraDataBlock):
    def __init__(self):
        ExtraDataBlock.__init__(self)
        self.folderid = GUID()
        self.offset = v_uint32()


class ShimBlock(ExtraDataBlock):
    def __init__(self):
        ExtraDataBlock.__init__(self)
        self.layername = v_zwstr()


class TrackerBlock(ExtraDataBlock):
    # See: https://winprotocoldoc.blob.core.windows.net/productionwindowsarchives/MS-DLTW/%5bMS-DLTW%5d.pdf
    def __init__(self):
        ExtraDataBlock.__init__(self)
        self.length = v_uint32()
        self.version = v_uint32()
        self.machineid = v_str(size=16)

        self.droidfile = GUID()
        self.droidvolume = GUID()

        self.droibirthfile = GUID()
        self.droibirthvolume = GUID()


class RootFolderItem(ItemBase):
    def __init__(self):
        ItemBase.__init__(self)

        self.sortidx = v_uint8()
        self.folderiden = GUID()

        # TODO: There's a pair of possible extension items here, but they're undocumented
        # even through extensive research by others. Skip for now.


class VolumeItem(ItemBase):
    def __init__(self):
        ItemBase.__init__(self)

    def pcb_iden(self):
        if self.iden & 0x1 == 0:
            self.flags = v_uint8()
        self.volume = v_str()


class FileEntryItem(ItemBase):
    def __init__(self):
        ItemBase.__init__(self)

    def pcb_iden(self):
        self.unk = v_uint8()
        self.filesize = v_uint32()
        self.lastmod = v_uint32()

        self.fileattrs = v_uint16()

        unicode = self.iden & 0x04 != 0
        if unicode:
            self.primary_name = v_zwstr()
        else:
            self.primary_name = v_zstr()

    def pcb_primary_name(self):
        # if this is an ascii string and we're not 16-bit aligned,
        # consume an extra padded byte to get us to 16 alignment
        unicode = self.iden & 0x04 != 0
        # include null terminator in length calculation
        if unicode == 0 and ((len(self.primary_name) + 1) * 8) % 16 != 0:
            self.padding = v_bytes(size=1)

        self.extsize = v_uint16()
        self.extver = v_uint16()
        self.extsig = v_uint32()
        self.extcreate = v_uint32()
        self.extaccess = v_uint32()
        self.extunk = v_uint16()

    def pcb_extver(self):
        if self.extver >= 7:
            self.o_unk = v_uint16()
            self.o_ref = v_uint64()
            self.o_wat = v_uint64()

        if self.extver >= 3:
            self.longstr_size = v_uint16()

        if self.extver >= 9:
            self.unk3 = v_uint32()

        if self.extver >= 8:
            self.unk4 = v_uint32()

        if self.extver >= 3:
            self.longname = v_zwstr()

    def pcb_longstr_size(self):
        if self.longstr_size > 0:
            if self.extver >= 7:
                self.localname = v_zwstr()
            elif self.extver >= 3:
                self.localname = v_str()

        if self.extver >= 3:
            self.firstextoffset = v_uint16()


class CommonNetworkRelativeLink(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.size = v_uint32()
        self.flags = v_uint32()

        self.netnameoffset = v_uint32()
        self.devicenameoffset = v_uint32()
        self.networkprovidertype = v_uint32()

    def pcb_networkprovidertype(self):
        if self.netnameoffset > 0x14:
            self.netnameoffsetunicode = v_wstr()
            self.devicenameoffsetunicode = v_wstr()

        self.netname = v_str()
        self.devicename = v_str()

        self.netnameunicode = v_wstr()
        self.devicenameunicode = v_wstr()


class LinkInfo(vstruct.VStruct):
    def __init__(self):
        vstruct.VStruct.__init__(self)

        self.size = v_uint32()
        self.headersize = v_uint32()
        self.flags = v_uint32()

        self.VolumeIdOffset = v_uint32()
        self.LocalBasePathOffset = v_uint32()
        self.CommonNetworkRelativeLinkOffset = v_uint32()
        self.CommonPathSuffixOffset = v_uint32()

    def pcb_CommonNetworkRelativeLinkOffset(self):
        if self.headersize >= 0x24:
            self.LocalBasePathOffsetUnicode = v_uint32()
            self.CommonPathSuffixOffsetUnicode = v_uint32()

        if self.flags & LinkInfoFlags.VolumeIDAndLocalBasePath:
            self.volumeid = VolumeID()
            self.localbasepath = v_zstr()

        if self.flags & LinkInfoFlags.CommonNetworkRelativeLinkAndPathSuffix:
            self.CommonNetworkRelativeLink = CommonNetworkRelativeLink()

        self.CommonPathSuffix = v_zstr()

        if self.flags & LinkInfoFlags.VolumeIDAndLocalBasePath and self.headersize >= 0x24:
            self.LocalBasePathUnicode = v_zwstr()
            self.CommonPathSuffixUnicode = v_zwstr()


class StringDataItem(vstruct.VStruct):
    def __init__(self, unicode=False):
        vstruct.VStruct.__init__(self)

        self.unicode = unicode
        self.count = v_uint16()

    def pcb_count(self):
        # The "string" mentioned is specifcally stated "MUST NOT be NULL TERMINATED"
        if self.unicode:
            self.data = v_bytes(size=self.count*2)
        else:
            self.data = v_bytes(size=self.count)


class Lnk:
    def __init__(self, byts):
        offset = 0
        self.header = LnkShellLinkHeader()
        self.header.vsParse(byts)

        self.items = None
        self.linkinfo = None

        # StringData items
        self.nameInfo = None
        self.relPath = None
        self.workingDir = None
        self.cmdargs = None
        self.iconLoc = None

        # ExtraData items
        self.console = None
        self.consolefe = None
        self.darwin = None
        self.envVar = None
        self.iconEnvData = None
        self.knownFolder = None
        self.propStore = None
        self.shimData = None
        self.specialFolder = None
        self.trackerData = None
        self.vistaIDList = None

        self.leftovers = None

        offset += len(self.header)
        unicode = self.header.linkflags & LinkFlags.IsUnicode != 0

        if self.header.linkflags & LinkFlags.HasLinkTargetIDList:
            self.items, offset = self._parseItemID(byts, offset)

        if self.header.linkflags & LinkFlags.HasLinkInfo:
            # LinkInfo
            self.linkinfo = LinkInfo()
            self.linkinfo.vsParse(byts[offset:])

            offset += self.linkinfo.size

        # NOTE: StringData isn't really one unifed structure.
        # It's a collection of things controlled by various flags that all happen to look the same.
        if self.header.linkflags & LinkFlags.HasName:
            self.nameInfo = StringDataItem(unicode=unicode)
            self.nameInfo.vsParse(byts[offset:])
            offset += len(self.nameInfo)

        if self.header.linkflags & LinkFlags.HasRelativePath:
            self.relPath = StringDataItem(unicode=unicode)
            self.relPath.vsParse(byts[offset:])
            offset += len(self.relPath)

        if self.header.linkflags & LinkFlags.HasWorkingDir:
            self.workingDir = StringDataItem(unicode=unicode)
            self.workingDir.vsParse(byts[offset:])
            offset += len(self.workingDir)

        if self.header.linkflags & LinkFlags.HasArguments:
            self.cmdargs = StringDataItem(unicode=unicode)
            self.cmdargs.vsParse(byts[offset:])
            offset += len(self.cmdargs)

        if self.header.linkflags & LinkFlags.HasIconLocation:
            self.iconLoc = StringDataItem(unicode=unicode)
            self.iconLoc.vsParse(byts[offset:])
            offset += len(self.iconLoc)

        # ExtraData
        while True:
            if offset >= len(byts) - 3:
                break

            size = v_uint32()
            size.vsParse(byts[offset:])
            if size < 0x4:
                # Terminal "block"
                offset += 4
                break

            item = ExtraDataBlock()
            item.vsParse(byts[offset:])
            if item.blocksig == ExtraDataSigs.PropertyStore:
                # The spec doesn't mention this at all, but this isn't a SerializedPropertyStore
                # it's a seriesi of SerializedPropertyStorage (yes those are different) structures
                # and you know to stop when the last has a size of 0, like a few of the other structures
                # in this LNK file
                try:
                    propStore = vstruct.VArray()
                    step = len(item)
                    while True:
                        ps = PropertyStorage()
                        ps.vsParse(byts[offset+step:])
                        propStore.vsAddElement(ps)
                        step += ps.size
                        if ps.size == 0:
                            break
                    self.propStore = propStore
                except Exception as exc:
                    logger.warning('PropertyStore failed to parse: %r', exc)
                offset += item.blocksize
                # offset += self.propStore.size + 8  # This gets us to some kinda alignment I think?
                continue
            elif item.blocksig == ExtraDataSigs.VistaAndAboveIDList:
                self.vistaIDList, offset = self._parseItemID(byts, offset)
                continue
            elif item.blocksig == ExtraDataSigs.Console:
                self.console = block = ConsoleBlock()
            elif item.blocksig == ExtraDataSigs.ConsoleFE:
                self.consolefe = block = ConsoleFEBlock()
            elif item.blocksig == ExtraDataSigs.Darwin:
                self.darwin = block = TargetBlock()
            elif item.blocksig == ExtraDataSigs.Environment:
                self.envVar = block = TargetBlock()
            elif item.blocksig == ExtraDataSigs.IconEnvironment:
                self.iconEnvData = block = TargetBlock()
            elif item.blocksig == ExtraDataSigs.KnownFolder:
                self.knownFolder = block = FolderBlock()
            elif item.blocksig == ExtraDataSigs.Shim:
                self.shimData = block = ShimBlock()
            elif item.blocksig == ExtraDataSigs.SpecialFolder:
                self.specialFolder = block = FolderBlock()
            elif item.blocksig == ExtraDataSigs.Tracker:
                self.trackerData = block = TrackerBlock()
            else:
                # More malicious ones will stash things here that aren't valid blocks, so
                # we can't really go any further if the block is invalid
                logger.warning(f'Unhandled ExtraData with Signature {hex(item.blocksig)} of size {item.blocksize}')
                offset += item.blocksize
                continue

            block.vsParse(byts[offset:])
            offset += item.blocksize

        if offset < len(byts):
            self.leftovers = byts[offset:]

    def _parseItemID(self, byts, offset):
        items = []
        idlistsize = v_uint16()
        idlistsize.vsParse(byts[offset:])
        offset += len(idlistsize)

        while True:
            base = ItemBase()
            base.vsParse(byts[offset:])

            size = base.size
            iden = base.iden
            if size == 0:
                offset += 2
                break
            elif iden == 0x1f:
                item = RootFolderItem()
                item.vsParse(byts[offset:])
                offset += item.size
            elif iden & 0x70 == 0x20:
                item = VolumeItem()
                item.vsParse(byts[offset:])
                offset += item.size
            elif iden & 0x70 == 0x30:
                item = FileEntryItem()
                item.vsParse(byts[offset:])
                offset += item.size
            else:
                logger.warning(f'ItemID: Got unhandled iden of {hex(iden)}?')
                offset += size
                continue

            items.append(item)

        return items, offset


def parseFromBytes(byts):
    return Lnk(byts)


if __name__ == '__main__':
    import sys
    with open(sys.argv[1], 'rb') as fd:
        parseFromBytes(fd.read())



'''
The envi.memcanvas module is the home of the base MemoryRenderer object and
MemoryCanvas objects.
'''

import sys
import logging
import functools
import traceback

import envi.symstore.resolver as e_resolv

logger = logging.getLogger(__name__)

class MemoryRenderer(object):
    """
    A top level object for all memory renderers
    """

    def rendSymbol(self, mcanv, va):
        """
        If there is a symbolic name for the current va, print it...
        """
        sym = mcanv.syms.getSymByAddr(va)
        if sym is not None:
            mcanv.addVaText("%s:\n" % repr(sym), va)

    def rendVa(self, mcanv, va):
        tag = mcanv.getVaTag(va)
        mcanv.addText("%.8x:" % va, tag=tag)

    def rendChars(self, mcanv, bytez):
        for b in bytez:
            bstr = "%.2x" % b
            if b < 0x20 or b > 0x7e:
                b = "."
            else:
                b = chr(b)
            mcanv.addNameText(b, bstr)

    def render(self, mcanv, va):
        """
        Render one "unit" and return the size you ate.
        mcanv will be a MemoryCanvas extender and va
        is the virtual address you are expected to render.
        """
        raise Exception("Implement render!")


class MemoryCanvas(object):
    """
    A memory canvas is a place where the textual representation
    of memory will be displayed. The methods implemented here show
    how a memory canvas which simply prints would be implemented.
    """
    def __init__(self, mem=None, syms=None):
        if mem is None:
            raise Exception("MemoryCanvas must include mem args")
        if syms is None:
            syms = e_resolv.SymbolResolver()
        self.mem = mem
        self.syms = syms
        self.currend = None
        self.renderers = {}
        self._canv_scrolled = False
        self._canv_navcallback = None

        # A few things for tracking renders.
        self._canv_beginva = None
        self._canv_endva = None
        self._canv_rendvas = []

    def setScrolledCanvas(self, scroll):
        self._canv_scrolled = scroll

    def write(self, msg):
        # So a canvas can act like simple standard out
        self.addText(msg)

    def setNavCallback(self, callback):
        '''
        Set a navigation "callback" that will be called with
        a memory expression as it's first argument anytime the
        canvas receives user input which desires nav...
        '''
        self._canv_navcallback = callback

    def getEndian(self):
        '''
        returns the endianness setting of the mem object
        '''
        return self.mem.getEndian()

    def setEndian(self, bigend):
        '''
        envi.ENDIAN_LSB or envi.ENDIAN_MSB
        passes the value to the mem.setEndian() accessor
        '''
        self.mem.setEndian(bigend)

    def addRenderer(self, name, rend):
        self.renderers[name] = rend
        self.currend = rend

    def getRenderer(self, name):
        return self.renderers.get(name)

    def getRendererNames(self):
        ret = list(self.renderers.keys())
        ret.sort()
        return ret

    def setRenderer(self, name):
        rend = self.renderers.get(name)
        if rend is None:
            raise Exception("Unknown renderer: %s" % name)
        self.currend = rend

    def getTag(self, typename):
        """
        Retrieve a non-named tag (doesn't highlight or do
        anything particularly special, but allows color
        by typename).
        """
        return None

    def getNameTag(self, name, typename='name'):
        """
        Retrieve a "tag" object for a name.  "Name" tags will
        (if possible) be highlighted in the rendered interface
        """
        return None  # No highlighting in plain text

    def getVaTag(self, va):
        """
        Retrieve a tag object suitable for showing that the text
        added with this tag should link through to the specified
        virtual address in the memory canvas.
        """
        return None  # No linking in plain text

    def addText(self, text, tag=None):
        """
        Add text to the canvas with a specified tag.

        NOTE: Implementors should probably check _canv_scrolled to
        decide if they should scroll to the end of the view...
        """
        sys.stdout.write(text)

    def addNameText(self, text, name=None, typename='name'):
        if name is None:
            name = bytes([ord(x) for x in text])
        else:
            name = bytes([ord(x) for x in name])
        tag = self.getNameTag(name, typename=typename)
        self.addText(text, tag=tag)

    def addVaText(self, text, va):
        tag = self.getVaTag(va)
        self.addText(text, tag=tag)

    def render(self, va, size, rend=None):
        raise Exception('Deprecated!  use renderMemory!')

    def clearCanvas(self, cb=None, sel=None):
        if cb is not None:
            cb(None)

    def _beginRenderMemory(self, va, size, rend):
        pass

    def _endRenderMemory(self, va, size, rend, cb=None, sel=None):
        if cb is not None:
            cb(None)

    def _beginRenderVa(self, va):
        pass

    def _endRenderVa(self, va):
        pass

    def _beginUpdateVas(self, valist):
        raise Exception("Default canvas can't update!")

    def _endUpdateVas(self):
        pass

    def _beginRenderAppend(self):
        raise Exception("Default canvas can't append!")

    def _endRenderAppend(self):
        pass

    def _beginRenderPrepend(self):
        raise Exception("Default canvas can't prepend!")

    def _endRenderPrepend(self):
        pass

    def _isRendered(self, va, maxva):
        '''
        Returns true if any part of the current render overlaps
        with the specified region.
        '''
        if self._canv_beginva is None:
            return False

        if self._canv_endva is None:
            return False

        if va > self._canv_endva:
            return False

        if maxva < self._canv_beginva:
            return False

        return True

    def _loc_helper(self, va):
        '''
        allows subclassess to make the starting VA make more contextual sense.
        '''
        return (va, 0)

    def renderMemoryUpdate(self, va, size, init=None, fini=None):

        maxva = va + size
        if not self._isRendered(va, maxva):
            return

        # Find the index of the first and last change
        iend = None
        ibegin = None
        for i, (rendva, rendsize) in enumerate(self._canv_rendvas):

            if ibegin is None and va <= rendva:
                ibegin = i

            if iend is None and maxva <= rendva:
                iend = i

            if ibegin is not None and iend is not None:
                break

        saved_last = self._canv_rendvas[iend:]
        saved_first = self._canv_rendvas[:ibegin]
        updatedvas = self._canv_rendvas[ibegin:iend]

        # We must actually start rendering from the beginning
        # of the first updated VA index
        startva = updatedvas[0][0]
        endva = self._canv_endva
        if saved_last:
            endva = saved_last[0][0]

        newrendvas = []

        self._beginUpdateVas(updatedvas, init)
        try:

            while startva < endva:
                self._beginRenderVa(startva)
                rsize = self.currend.render(self, startva)
                newrendvas.append((startva, rsize))
                self._endRenderVa(startva)
                startva += rsize

        except Exception:
            s = traceback.format_exc()
            self.addText("\nException At %s: %s\n" % (hex(va), s))

        self._canv_rendvas = saved_first + newrendvas + saved_last

        self._endUpdateVas(fini)

    def renderMemoryPrepend(self, size, cb=None):
        firstva, firstsize = self._canv_rendvas[0]

        va, szdiff = self._loc_helper(firstva - size)
        size += szdiff

        self._beginRenderPrepend()

        savedrendvas = self._canv_rendvas
        self._canv_rendvas = []
        self._canv_beginva = va

        rend = self.currend

        try:

            while va < firstva:
                self._beginRenderVa(va)
                rsize = rend.render(self, va)
                self._canv_rendvas.append((va, rsize))
                self._endRenderVa(va)
                va += rsize

            self._canv_rendvas.extend(savedrendvas)

        except Exception:
            s = traceback.format_exc()
            self.addText("\nException At %s: %s\n" % (hex(va), s))

        self._endRenderPrepend(cb)

    def renderMemoryAppend(self, size, cb=None):
        lastva, lastsize = self._canv_rendvas[-1]

        va = lastva + lastsize

        self._beginRenderAppend()

        rend = self.currend
        try:
            maxva = va + size
            while va < maxva:
                self._beginRenderVa(va)
                rsize = rend.render(self, va)
                self._canv_rendvas.append((va, rsize))
                self._endRenderVa(va)
                va += rsize

            self._canv_endva = maxva

        except Exception:
            s = traceback.format_exc()
            self.addText("\nException At %s: %s\n" % (hex(va), s))

        self._endRenderAppend(cb)

    def _canvasCleared(self, cb, sel, data):
        va = self._canv_beginva
        maxva = self._canv_endva
        size = maxva - va
        rend = self.currend
        # A callback for "bulk" rendering (let the canvas cache...)
        self._beginRenderMemory(va, size, rend)
        try:
            maxva = va + size
            while va < maxva:

                self._beginRenderVa(va)
                try:
                    rsize = rend.render(self, va)
                    self._canv_rendvas.append((va, rsize))
                    self._endRenderVa(va)
                    va += rsize
                except Exception as e:
                    logger.error(traceback.format_exc())
                    self.addText("\nRender Exception At %s: %s\n" % (hex(va), str(e)))
                    self._endRenderVa(va)
                    break

        except Exception as e:
            self.addText("\nException At %s: %s\n" % (hex(va), str(e)))

        # Canvas callback for render completion (or error...)
        self._endRenderMemory(va, size, rend, cb, sel)

    def renderMemory(self, va, size, rend=None, cb=None, clear=True, sel=None):
        if sel is None:
            sel = '#memcanvas'
        # Set our canvas render tracking variables.
        self._canv_beginva = va
        self._canv_endva = va + size
        self._canv_rendvas = []
        if rend is None:
            rend = self.currend
        self.currend = rend

        clearcb = functools.partial(self._canvasCleared, cb, sel)
        # if this is not a "scrolled" canvas, clear it.
        if not self._canv_scrolled and clear:
            self.clearCanvas(clearcb, sel)
        else:
            clearcb(None)


class StringMemoryCanvas(MemoryCanvas):

    def __init__(self, mem, syms=None):
        MemoryCanvas.__init__(self, mem, syms=syms)
        self.strval = ''

        # we perform manual clearing of the canvas.
        # we don't want it cleared every renderMemory call.
        self.setScrolledCanvas(True)

    def clearCanvas(self, cb=None, sel=None):
        self.strval = ''

    def addText(self, text, tag=None):
        self.strval += text

    def __str__(self):
        return self.strval


class FileBackedMemoryCanvas(StringMemoryCanvas):
    '''
    FileBackedMemoryCanvas writes output to a file of your choosing.  It writes
    out the data to the file after `rollwindow` number of calls to `addText()`.

    NOTE: when done writing to the canvas, be certain to call `flush()` or you
            will lose data.
    '''
    def __init__(self, filename, mem, syms=None, append=True, rollwindow=100):
        StringMemoryCanvas.__init__(self, mem, syms)
        if append:
            self.fd = open(filename, 'a', encoding='utf-8')
        else:
            self.fd = open(filename, 'w', encoding='utf-8')

        self.rollwindow = rollwindow
        self.linecount = 0

    def addText(self, text, tag=None):
        super().addText(text, tag)

        self.linecount += 1
        if self.linecount % self.rollwindow == 0:
            self.flush()

    def flush(self):
        self.fd.write(str(self))
        self.clearCanvas()
        self.linecount = 0

    def __del__(self):
        self.flush()
        self.fd.close()

class CanvasMethodProxy(object):
    '''
    Target for teecanvas.
    '''
    def __init__(self, canvases, name):
        self.canvases = canvases
        self.name = name

    def __call__(self, *args, **kwargs):
        for canvas in self.canvases:
            try:
                attr = getattr(canvas, self.name)
                attr(*args, **kwargs)
            except Exception as e:
                logger.warning("Error: %r", e, exc_info=1)

    def __getattr__(self, name):
        obj = getattr(self.canvases[0], self.name)
        return getattr(obj, name)


class TeeCanvas(object):
    '''
    Replaces the canvas on an object (temporarily) with a proxy canvas that
    forwards requests to other canvases.

    Example usage:
    with TeeCanvas(self, (self.canvas, canvas2)) as tc:
        self.onecmd(command)
    '''
    def __init__(self, target, canvases):
        self.target = target
        self.ocanvas = None
        self.canvases = canvases

    def __getattr__(self, name):
        return CanvasMethodProxy(self.canvases, name)

    def __enter__(self):
        '''
        replace the canvas of the target with ourselves.
        '''
        self.ocanvas = self.target.canvas
        self.target.canvas = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        '''
        restore the canvas of the target.
        '''
        self.target.canvas = self.ocanvas



"""Locate any xrefs/names to undefined locations"""

import envi.common as e_common
import vivisect.const as v_const

columns = (("Bytes", str), ("Name", str))


def report(vw):
    res = {}

    for fromva, tova, reftype, rflags in vw.getXrefs():
        if vw.getLocation(tova, range=True) is None:
            rname = v_const.ref_type_names.get(reftype, "Unknown")
            sname = "Unknown"
            seg = vw.getSegment(tova)
            if seg is not None:
                sname = seg[v_const.SEG_NAME]
            try:
                b = e_common.hexify(vw.readMemory(tova, 8))
            except Exception as e:
                b = str(e)
            res[tova] = (b, "%s ref from 0x%x (%s)" % (rname, fromva, sname))

    for va, name in vw.getNames():
        if vw.getLocation(va) is None:
            try:
                b = e_common.hexify(vw.readMemory(tova, 8))
            except Exception as e:
                b = str(e)
            res[va] = (b, name)

    return res




import vstruct
from vstruct.primitives import *

# Mapped at 0xe000e000 in arm7
class SCS(vstruct.VStruct):

    def __init__(self):
        vstruct.VStruct.__init__(self)
        self.MasterControl           = v_uint32()
        self.InterruptControllerType = v_uint32()
        self.AuxiliaryControl        = v_uint32()
        self.ReservedSpace           = v_bytes(3320)
        self.CPUIDBase               = v_uint32()
        self.InterruptControlState   = v_uint32()
        self.VectorTableOffset       = v_uint32()
        self.AppInterr_ResetControl  = v_uint32()
        self.SystemControl           = v_uint32()
        self.ConfigurationControl    = v_uint32()



"""
A byte and mask based decision engine for creating byte
sequences (and potential comparison masks) for general purpose
signature matching.

Currently used by vivisect function entry sig db and others.
"""


class SignatureTree:
    """
    A byte based decision tree which uses all the RAMs but is
    really fast....

    Signatures consist of a byte sequence and an optional mask
    sequence.  If present each mask byte is used to logical and
    the byte being compared before comparison.  This allows the
    creation of signatures which have parts of the sig generalized.

    FIXME allow sigs to have a reliability rating
    FIXME allow sig nodes to store depth and truncate the tree early (and then mask the rest)
    """

    def __init__(self):
        # each node in the tree is a tuple consisting of the depth we're at,
        # the signatures in this particular subtree, and the list of subtree nodes
        self.basenode = (0, [], [None] * 256, [])
        self.sigs = {}  # track duplicates

    def _addChoice(self, siginfo, node):

        todo = [(node, siginfo),]

        # Recursion is for the weak minded.
        while len(todo):

            node, siginfo = todo.pop()
            depth, sigs, choices, term = node
            bytes, masks, o = siginfo
            siglen = len(sigs)
            if len(bytes) > depth:
                sigs.append(siginfo)
            else:
                term.append(siginfo)
                continue
            # If one sig is [85, 139, 236] and another is [85, 139, 236, 232, 144], then
            # we're gonna hit an IndexException without this check
            if siglen == 0:
                pass # we just don't want the "else" here, if we're the only
                # one on this node, just let it ride.

            elif siglen == 1:
                # If it has one already, we *both* need to add another level
                # (because if it is the only one, it thought it was last choice)
                for sig in sigs:
                    chval = sig[0][depth] # get the choice byte from siginfo
                    nnode = self._getNode(depth, choices, chval)
                    todo.append((nnode, sig))

            else: # This is already a choice node, keep on choosing...
                chval = bytes[depth]
                nnode = self._getNode(depth, choices, chval)
                todo.append((nnode, siginfo))

    def _getNode(self, depth, choices, choice):
        # Chose, (and or initialize) a sub node
        nnode = choices[choice]
        if nnode is None:
            nnode = (depth+1, [], [None] * 256, [])
            choices[choice] = nnode
        return nnode

    def addSignature(self, bytes, masks=None, val=None):
        """
        Add a signature to the search tree.  If masks goes unspecified, it will be
        assumed to be all ones (\\xff * len(bytes)).

        Additionally, you may specify "val" as the object to get back with
        getSignature().
        """
        # FIXME perhaps make masks None on all ff's
        if masks is None:
            masks = b'\xff' * len(bytes)

        if val is None:
            val = True

        # Detect and skip duplicate additions...
        bytekey = bytes + masks
        if self.sigs.get(bytekey) is not None:
            return

        self.sigs[bytekey] = True

        siginfo = (bytes, masks, val)
        self._addChoice(siginfo, self.basenode)

    def isSignature(self, bytes, offset=0):
        return self.getSignature(bytes, offset=offset) is not None

    def getSignature(self, bytes, offset=0):
        matches = []
        node = self.basenode
        while True:
            depth, sigs, choices, term = node
            matches.extend(term)
            # Once we get down to one sig, there are no more branches,
            # just check the byte sequence.
            if len(sigs) == 1:
                sbytes, smasks, sobj = sigs[0]
                is_match = True
                for i in range(depth, len(sbytes)):
                    realoff = offset + i
                    # we still have pieces of the signature left to match, but bytes wasn't long enough
                    if realoff >= len(bytes):
                        is_match = False
                        break
                    masked = bytes[realoff] & smasks[i]
                    if masked != sbytes[i]:
                        is_match = False
                        break
                if is_match:
                    matches.append(sigs[0])
                break

            # There are still more choices, keep branching.
            node = None # Lets go find a new one
            for sig in sigs:
                sbytes, smasks, sobj = sig
                if offset+depth >= len(bytes):
                    continue
                # we've reached the end of this signature, so we're just going to mask the rest
                masked = bytes[offset+depth] & smasks[depth]
                if sbytes[depth] == masked: # We have a winner!
                    # FIXME find the *best* winner! (because of masking)
                    node = choices[masked]
                    break

            # We failed to make our next choice
            if node is None:
                break
        if len(matches) == 0:
            return None
        return sorted(matches, key=lambda m: len(m[0]), reverse=True)[0][2]



import json
import logging

logger = logging.getLogger(__name__)

def compat_isNone(state):
    # WTF! (QByteArray is None) is True!
    if state is None:
        return True

    return not len(state)

class SaveableWidget(object):
    '''
    Inherited by widgets that want to save and restore settings.

    Implement vqGetSaveState/vqSetSaveState.
    '''
    def vqSaveState(self, settings, name, stub=''):
        state = self.vqGetSaveState()
        settings.setValue(stub+name, json.dumps(state))

    def vqRestoreState(self, settings, name, stub=''):
        qstate = settings.value(stub+name)
        if qstate is None:
            return

        try:
            state = json.loads(str(qstate))
            self.vqSetSaveState(state)
        except Exception as e:
            logger.warning('failed to restore %s: %s', name, e)

    def vqGetSaveState(self):
        return None

    def vqSetSaveState(self, state):
        return None



import logging

import envi.common as e_common

import vivisect.parsers as viv_parsers

import vstruct.defs.macho as vs_macho
import vstruct.defs.macho.fat as vsm_fat
import vstruct.defs.macho.const as vsm_const

logger = logging.getLogger(__name__)


def parseFile(vw, filename, baseaddr=None):
    with open(filename, 'rb') as f:
        fbytes = f.read()
    return _loadMacho(vw, fbytes, filename=filename, baseaddr=baseaddr)


def parseBytes(vw, filebytes, baseaddr=None):
    return _loadMacho(vw, filebytes, baseaddr=baseaddr)


def parseMemory(vw, memobj, baseaddr):
    byts = memobj.read()
    return _loadMacho(vw, byts, baseaddr=baseaddr)


archcalls = {
    'i386': 'cdecl',
    'amd64': 'sysvamd64call',
    'arm': 'armcall',
    'thumb': 'armcall',
    'thumb16': 'armcall',
}

arch_translation = {
    'powerpc': 'ppc-server',
}

'''
The various magic byte strings are as follows:
cefaedfe: Mach-O Little Endian (32-bit)
cffaedfe: Mach-O Little Endian (64-bit)
feedface: Mach-O Big Endian (32-bit)
feedfacf: Mach-O Big Endian (64-bit)
cafebabe: Universal Binary Big Endian. These fat binaries are archives that can include binaries for multiple architectures, but typically contain PowerPC and Intel x86.
'''

def checkFatMagicAndTrunc(vw, filebytes):
    archhdr = None
    fatarch = vw.config.viv.parsers.macho.fatarch

    archlist = []

    offset = 0
    fat = vs_macho.fat_header()
    offset = fat.vsParse(filebytes, offset=offset)
    
    if fat.magic in vsm_const.FAT_64:
        archcon = vsm_fat.fat_arch_64
    else:
        archcon = vsm_fat.fat_arch

    for i in range(fat.nfat_arch):
        ar = vs_macho.fat_arch()
        offset = ar.vsParse(filebytes, offset=offset)
        archname = vs_macho.mach_cpu_names.get(ar.cputype)
        if archname == fatarch:
            archhdr = ar
            break
        archlist.append((archname, ar))

    if not archhdr:
        # If we don't have a specified arch, exception!
        vw.vprint('Mach-O Fat Binary Architectures:')
        for archname, ar in archlist:
            vw.vprint('0x%.8x 0x%.8x %s' % (ar.offset, ar.size, archname))
        raise Exception('Mach-O Fat Binary: Please specify arch with -O viv.parsers.macho.fatarch="<archname>"')

    filebytes = filebytes[archhdr.offset:archhdr.offset+archhdr.size]

    return filebytes

def _loadMacho(vw, filebytes, filename=None, baseaddr=None):

    hdr = e_common.hexify(filebytes[:4])

    # Check for the FAT binary magic...
    if hdr in ('cafebabe', 'bebafeca'):
        filebytes = checkFatMagicAndTrunc(vw, filebytes)

    # Instantiate the parser wrapper and parse bytes
    macho = vs_macho.mach_o()
    macho.vsParse(filebytes)
    # find the lowest loadable address, then get an offset so we can apply it later
    fakebase, size = getMemBaseAndSize(vw, macho, baseaddr=baseaddr)
    offset = baseaddr - fakebase
    logger.debug('initial file baseva: 0x%x  size: 0x%x (address offset: 0x%x)', fakebase, size, offset)

    # Determine base address to load into
    # We fake them to *much* higher than norm so pointer tests do better...
    # 1) baseaddr arg
    # 2) vw.config.viv.parsers.macho.baseaddr if non-zero
    # 3) vw.findFreeMemoryBlock()  if baseaddr is zero/None or already exists in another map (collision)

    if baseaddr is None:
        baseaddr = vw.config.viv.parsers.macho.baseaddr

    if vw.isValidPointer(baseaddr):
        logger.info("baseaddr (0x%x) is in use.  Finding appropriate address base.", baseaddr)
        baseaddr = vw.findFreeMemoryBlock(size, fakebase)
        logger.debug("loading %r (size: 0x%x) at 0x%x", filename, size, baseaddr)

    if filename is None:
        filename = 'macho_%.8x' % baseaddr  # FIXME more than one!

    # grab md5 and sha256 hashes before we modify the bytes
    fhash = viv_parsers.md5Bytes(filebytes)
    sha256 = viv_parsers.sha256Bytes(filebytes)


    arch = vs_macho.mach_cpu_names.get(macho.mach_header.cputype)
    if arch is None:
        raise Exception('Unknown MACH-O arch: %.8x' % macho.mach_header.cputype)

    vwarch = arch_translation.get(arch, arch)

    # Setup arch/plat/fmt
    vw.setMeta('Architecture', vwarch)
    vw.setMeta("Platform", "Darwin")
    vw.setMeta("Format", "macho")

    vw.setMeta('DefaultCall', archcalls.get(arch, 'unknown'))

    # Add the file entry
    fname = vw.addFile(filename, baseaddr, fhash)
    vw.setFileMeta(fname, 'sha256', sha256)

    logger.debug("loading memory maps...")
    # Add the memory maps and segments from the macho definition
    for segname, rva, perms, segbytes in macho.getSegments():
        logger.debug("map: %r rva:0x%x perms:0x%x len: 0x%x", segname, rva, perms, len(segbytes))
        segbase = rva + offset  # not baseaddr, since it already has baseaddr applied
        vw.addMemoryMap(segbase, perms, fname, segbytes)
        vw.addSegment(segbase, len(segbytes), segname, fname)

    # Add the library dependancies
    for libname in macho.getLibDeps():
        logger.debug("deplib: %r", libname)
        # FIXME hack....
        libname = libname.split('/')[-1]
        libname = libname.split('.')[0].lower()
        vw.addLibraryDependancy(libname)

    logger.debug("memoryMaps: \n%r", '\n'.join(["0x%x, 0x%x, 0x%x, %r" % (w,x,y,z) for w,x,y,z in vw.getMemoryMaps()]))
    # Add Mach-O structures
    structinfo, cmdinfo = macho.getStructureInfo()
    for sname, va in structinfo:
        try:
            logger.debug("Applying struct %r at 0x%x", sname, va+offset)
            vw.makeStructure(va + offset, sname)
        except Exception as e:
            logger.warning("Error: %r", e)

    for va, cmdname in cmdinfo:
        vw.makeName(va, "macho_" + cmdname)


    for soff, va in macho.getEntryPoints():
        va += baseaddr
        if vw.isValidPointer(va):
            vw.addEntryPoint(va)

        ptr = baseaddr + soff
        logger.debug("adding entrypoint: 0x%x (off: 0x%x/0x%x)", va, ptr, soff) 
        vw.makePointer(ptr, follow=False)

    return fname

def getMemBaseAndSize(vw, macho, baseaddr=None):
    '''
    Returns the default baseaddr and memory size required to load the file
    '''
    if baseaddr is None:
        baseaddr = vw.config.viv.parsers.macho.baseaddr

    memmaps = macho.getSegments()
    baseaddr = 0xffffffffffffffffffffffff
    topmem = 0

    for mname, mapva, mperms, mbytes in memmaps:
        if mapva < baseaddr:
            baseaddr = mapva
        endva = mapva + len(mbytes)
        if endva > topmem:
            topmem = endva

    size = topmem - baseaddr
    logger.debug("getMemBaseAndSize() => 0x%x:0x%x", baseaddr, size)
    return baseaddr, size



'''
The envi.qt.html module contains the HTML template and javascript
code used by the renderers (which are based on QtWebEngine)
'''

template = '''
<!DOCTYPE html>
<html id="mainhtml">
<style type="text/css">

body {
    color: #00ff00;
    background-color: #000000;
    white-space: pre;
    font: 10pt Monospace;
}

div.memcanvas {
    color: #00ff00;
    background-color: #000000;
}

div.codeblock {
    color: #00ff00;
    background-color: #000000;
    border: 2px solid #00ff00;
    display: inline-block;
}

div.codeblock:hover {
    border: 2px solid #ff0000;
}

.envi-va {
    color: #4040ff;
    background-color: #000000;
}

.envi-va-selected {
    color: #000000;
    background-color: #4040ff;
}

.envi-va:hover {
    font-weight: 900;
}

.envi-name {
    color: #00ff00;
    background-color: #000000;
}

.envi-name-selected {
    color: #000000;
    background-color: #00ff00;
}

.envi-registers {
    color: #ff0000;
    background-color: #000000;
}

.envi-registers-selected {
    color: #000000;
    background-color: #ff0000;
}

.envi-mnemonic {
    color: #ffff00;
    background-color: #000000;
}

.envi-mnemonic-selected {
    color: #000000;
    background-color: #ffff00;
}

.envi-funcflags {
    color: #ff0088;
    background-color: #000000;
}

.envi-funcflags-selected {
    color: #ff0088;
    background-color: #000000;
}

</style>

<style type="text/css" id="cmapstyle">
</style>

<head>
<script src="qrc:///qtwebchannel/qwebchannel.js"></script>
</head>
<script language="javascript">
{{{jquery}}}

var selclass = "name";
function nameclick(elem) {
    var elem = $(elem);
    var tagval = elem.attr('envival');
    var tagname = elem.attr('envitag');
    $("."+selclass).removeClass(selclass);
    selclass = "envi-" + tagname + "-selected";
    var newclass = "envi-" + tagname + "-" + tagval;
    $("." + selclass).removeClass(selclass);
    $("." + newclass).addClass(selclass);

    console.log("click name " + tagname + " : " + tagval);
    vnav._jsSetCurTag(tagname, tagval);
}

var curva = null;
var vnav = null;

document.addEventListener("DOMContentLoaded", function () {
    webc = new QWebChannel(qt.webChannelTransport, function (channel) {
        vnav = channel.objects.vnav;
    });
});



function vaclick(elem) {
    var elem = $(elem);
    var vastr = elem.attr("va");
    selectva(vastr);
}

function selectva(vastr) {
    var vaselect = ".envi-va-" + vastr;
    $(".envi-va-selected").removeClass("envi-va-selected");
    $(vaselect).addClass("envi-va-selected");
    vnav._jsSetCurVa(vastr);
}

function vagoto(elem) {
    vnav._jsGotoExpr($(elem).attr('va'))
}

function scrolltoid(name) {

    var elem = document.getElementById(name);

    if (elem != null) {
        var elemrect = elem.getBoundingClientRect();
        if ( elemrect.top < 0 ) {
            elem.scrollIntoView(true);
        }

        if ( elemrect.bottom > window.innerHeight ) {
            elem.scrollIntoView(false);
        }
    } else {
        console.log("Failed to find element");
    }
}

svgns = "http://www.w3.org/2000/svg";

function createSvgElement(ename, attrs) {
    var elem = document.createElementNS(svgns, ename);
    for (var aname in attrs) {
        elem.setAttribute(aname, attrs[aname]);
    }
    return elem
}

function svgwoot(parentid, svgid, width, height) {

    var elem = document.getElementById(parentid);

    var svgelem = createSvgElement("svg", { "height":height.toString(), "width":width.toString() })
    svgelem.setAttribute("id", svgid);

    elem.appendChild(svgelem);
}

function addSvgForeignObject(svgid, foid, width, height) {
    var foattrs = {
        "class":"node",
        "id":foid,
        "width":width,
        "height":height
    };

    var foelem = createSvgElement("foreignObject", foattrs);

    var svgelem = document.getElementById(svgid);
    svgelem.appendChild(foelem);
}

function addSvgForeignHtmlElement(foid, htmlid) {

    var foelem = document.getElementById(foid);
    var htmlelem = document.getElementById(htmlid);
    htmlelem.parentNode.removeChild(htmlelem);

    //foelem.appendChild(htmlid);

    var newbody = document.createElement("body");
    newbody.setAttribute("xmlns", "http://www.w3.org/1999/xhtml");
    newbody.appendChild( htmlelem );

    foelem.appendChild(newbody);
}

function moveSvgElement(elemid, xpos, ypos) {
    var elem = document.getElementById(elemid);
    elem.setAttribute("x", xpos);
    elem.setAttribute("y", ypos);
}

function plineover(pline) {
    pline.setAttribute("style", "fill:none;stroke:yellow;stroke-width:2")
}

function plineout(pline) {
    pline.setAttribute("style", "fill:none;stroke:green;stroke-width:2")
}

function drawSvgLine(svgid, lineid, points) {
    var plineattrs = {
        "id":lineid,
        "points":points,
        "style":"fill:none;stroke:green;stroke-width:2",
        "onmouseover":"plineover(this)",
        "onmouseout":"plineout(this)",
    };

    var lelem = createSvgElement("polyline", plineattrs);
    var svgelem = document.getElementById(svgid);

    //var rule = "polyline." + lineclass + ":hover { stroke: red; }";
    //document.styleSheets[0].insertRule(rule, 0);

    svgelem.appendChild(lelem);
}

</script>

<body id="vbody" width="999px"><div class="memcanvas" id="memcanvas"></div></body>

</html>
'''
